@@ -5,10 +5,16 @@ Module that provides functionality to determine if a CPA "matches" a CPP.  This
 case if it can be established that the Party in the CPP matches either the Party or the
 CounterParty in the CPA and that the service specifications match accordingly.
 
+TODO:
+-  Inline channel features prior to match
+-  Process defaults prior to match
+
 """
 
 import logging, traceback, re, lxml, isodate
 
+from unify import lookup_party_identifiers
+
 from schema import cppa3_content_model
 
 _NSMAP = {'cppa': 'http://docs.oasis-open.org/ebcore/ns/cppa/v3.0',
@@ -28,12 +34,15 @@ class MatchException(Exception):
 
 def match(cpp, cpa, handle_defaults=False):
     """
-    @param cpp:
-    @param cpa:
-    @return: a Boolean indication of the success of the match
+    @param cpp: a CPP that the CPA is to be matched against
+    @param cpa: a CPA that may or may not be derived from the CPP.
+    @return: the name of the party for which the match was found
+    A MatchException is raised if no match could be established
     """
     party_name, direction, this_role_element, other_role_element = match_parties(cpp, cpa)
-    match_service_specifications(cpp, cpa, direction, this_role_element, other_role_element)
+    other_parties = cpa_other_party_ids(cpa, direction)
+    acl_check(cpp, cpp, other_parties)
+    match_service_specifications(cpp, cpa, direction, this_role_element, other_role_element, other_parties)
     return party_name
 
 def match_parties(cpp, cpa):
@@ -44,7 +53,7 @@ def match_parties(cpp, cpa):
                                                namespaces=_NSMAP)[0])
         direction = lambda x: x
         this_role_element, other_role_element, id_position = 'cppa:PartyRole', 'cppa:CounterPartyRole', 0
-    except:
+    except MatchException:
         party_name = match_partyinfo(cpp.xpath('cppa:PartyInfo',
                                                namespaces=_NSMAP)[0],
                                      cpa.xpath('cppa:CounterPartyInfo',
@@ -62,23 +71,23 @@ def match_partyinfo(cpp_partyinfo, cpa_partyinfo):
         'cppa:PartyName/text()', namespaces=_NSMAP
     )[0]
     if cpa_party_name != cpp_party_name:
-        raise MatchException('Mismatch in party name: {} versus {}'.format(cpp_party_name,
+        raise_match_exception('Mismatch in party name: {} versus {}'.format(cpp_party_name,
                                                                            cpa_party_name))
     # All PartyIDs in the CPA PartyInfo must be present in the CPP PartyInfo
     #for party_id in cpp_partyinfo.xpath('cppa:PartyID')
     for party_id in cpa_partyinfo.xpath(
-        'cppa:PartyID', namespaces=_NSMAP
+        'cppa:PartyId', namespaces=_NSMAP
     ):
         party_id_value = party_id.text
         if party_id.get('type') == None:
-            cpp_partyid_list = cpp_partyinfo.xpath('cppa:PartyID/text()="{}"'.format(party_id_value),
+            cpp_partyid_list = cpp_partyinfo.xpath('cppa:PartyId/text()="{}"'.format(party_id_value),
                                                    namespaces=_NSMAP)
         else:
-            cpp_partyid_list = cpp_partyinfo.xpath('cppa:PartyID[@type="{}"]/text()="{}"'.format(party_id.get('type'),
-                                                                                                 party_id_value))
+            cpp_partyid_list = cpp_partyinfo.xpath('cppa:PartyId[@type="{}"]/text()="{}"'.format(party_id.get('type'),
+                                                                                                 party_id_value),
+                                                   namespaces=_NSMAP)
         if not cpp_partyid_list:
-            raise MatchException('Party ID {} not found in CPA'.format(party_id_value))
-
+            raise_match_exception('Party Id {} not found in CPA'.format(party_id_value))
 
     return cpp_party_name
 
@@ -100,7 +109,7 @@ def match_profileinfo(cpp, cpa, id_position):
         cpp_exp_date = isodate.parse_datetime(cpp_expirations[0].text)
         cpa_exp_date = isodate.parse_datetime(cpa_expirations[0].text)
         if cpa_exp_date > cpp_exp_date:
-            raise MatchException('Expiration {} must precede {}'.format(cpp_expirations[0].text,
+            raise_match_exception('Expiration {} must precede {}'.format(cpp_expirations[0].text,
                                                                         cpa_expirations[0].text))
     cpp_activations = cpp.xpath('cppa:ProfileInfo/cppa:ActivationDate',
                                    namespaces=_NSMAP)
@@ -110,12 +119,10 @@ def match_profileinfo(cpp, cpa, id_position):
         cpp_act_date = isodate.parse_datetime(cpp_activations[0].text)
         cpa_act_date = isodate.parse_datetime(cpa_activations[0].text)
         if cpa_act_date < cpp_act_date:
-            raise MatchException('Activation {} must precede {}'.format(cpp_activations[0].text,
+            raise_match_exception('Activation {} must precede {}'.format(cpp_activations[0].text,
                                                                         cpa_activations[0].text))
 
-
-
-def match_service_specifications(cpp, cpa, direction, this_role_element, other_role_element):
+def match_service_specifications(cpp, cpa, direction, this_role_element, other_role_element, other_parties):
     for cpa_service_specification in cpa.xpath('cppa:ServiceSpecification', namespaces=_NSMAP):
         party_role = cpa_service_specification.xpath('{}/@name'.format(this_role_element), namespaces=_NSMAP)[0]
         counterparty_role= cpa_service_specification.xpath('{}/@name'.format(other_role_element), namespaces=_NSMAP)[0]
@@ -123,27 +130,39 @@ def match_service_specifications(cpp, cpa, direction, this_role_element, other_r
                                                                  this_role_element,
                                                                  counterparty_role,
                                                                  other_role_element))
-        cpp_service_specification = cpp.xpath(
+        cpp_service_specification_list = cpp.xpath(
             'cppa:ServiceSpecification[cppa:PartyRole/@name="{}" and cppa:CounterPartyRole/@name="{}"]'.format(party_role,
                                                                                                                counterparty_role),
             namespaces=_NSMAP
-        )[0]
-        for cpa_service_binding in cpa_service_specification.xpath(
-            'cppa:ServiceBinding', namespaces=_NSMAP
-        ):
-            match_cpa_service_binding(cpp, cpp_service_specification, cpa, cpa_service_binding,
-                                      party_role, counterparty_role, direction)
+        )
+        if len(cpp_service_specification_list) > 0:
+            cpp_service_specification = cpp_service_specification_list[0]
+            for cpa_service_binding in cpa_service_specification.xpath(
+                'cppa:ServiceBinding', namespaces=_NSMAP
+            ):
+                match_cpa_service_binding(cpp, cpp_service_specification, cpa, cpa_service_binding,
+                                          party_role, counterparty_role, direction, other_parties)
+        else:
+            raise MatchException('No Service Specification for Role Pair {}, {}'.format(party_role,
+                                                                                        counterparty_role))
 
 def match_cpa_service_binding(cpp, cpp_service_specification, cpa, cpa_service_binding,
-                              party_role, counterparty_role, direction):
+                              party_role, counterparty_role, direction, other_parties):
     cpa_service_element = cpa_service_binding.xpath('cppa:Service', namespaces=_NSMAP)[0]
     cpa_service = cpa_service_element.text
     cpa_service_type = cpa_service_element.get('type')
+    acl_check(cpp_service_specification, cpp, other_parties)
+
     logging.info('Matching service binding for {}, {}, roles {}, {}'.format(cpa_service, cpa_service_type,
                                                                             party_role, counterparty_role))
-    cpp_service_binding = cpp_service_specification.xpath(service_xpath(cpa_service, cpa_service_type),
-                                                          namespaces=_NSMAP)[0]
-    match_service_bindings(cpp, cpp_service_binding, cpa, cpa_service_binding, cpa_service, direction)
+    cpp_service_binding_list = cpp_service_specification.xpath(service_xpath(cpa_service, cpa_service_type),
+                                                          namespaces=_NSMAP)
+    if len(cpp_service_binding_list) > 0:
+        cpp_service_binding = cpp_service_binding_list[0]
+        match_service_bindings(cpp, cpp_service_binding, cpa, cpa_service_binding, cpa_service, direction,
+                               other_parties)
+    else:
+        raise MatchException('No service {} {} found in CPP'.format(cpa_service, cpa_service_type))
 
 
 def service_xpath(value, type):
@@ -152,59 +171,137 @@ def service_xpath(value, type):
     else:
         return 'cppa:ServiceBinding[cppa:Service/text()="{}"]'.format(value)
 
-def match_service_bindings(cpp, cpp_service_binding, cpa, cpa_service_binding, cpa_service, direction):
+def match_service_bindings(cpp, cpp_service_binding, cpa, cpa_service_binding, cpa_service, direction,
+                           other_parties):
+    acl_check(cpp_service_binding, cpp, other_parties)
     for cpa_action_binding in cpa_service_binding.xpath(
         'cppa:ActionBinding', namespaces=_NSMAP
     ):
         cpa_action = cpa_action_binding.get('action')
         cpa_direction = cpa_action_binding.get('sendOrReceive')
         logging.info('For {}, checking CPA action {} ({})'.format(cpa_service, cpa_action, cpa_direction))
-        cpp_action_binding = cpp_service_binding.xpath(
+        cpp_action_binding_list = cpp_service_binding.xpath(
             'cppa:ActionBinding[@action="{}" and @sendOrReceive="{}"]'.format(cpa_action, direction(cpa_direction)),
             namespaces=_NSMAP
-        )[0]
-        match_action_bindings(cpp, cpp_action_binding, cpa, cpa_action_binding)
+        )
+        if len(cpp_action_binding_list) > 0:
+            cpp_action_binding = cpp_action_binding_list[0]
+            match_action_bindings(cpp, cpp_action_binding, cpa, cpa_action_binding, direction, other_parties)
+        else:
+            raise MatchException('No action {} for service {} in CPP'.format(cpa_action, cpa_service))
     # @todo also do the check that all actions that the CPP considers required are present in the CPA
 
-def match_action_bindings(cpp, cpp_action_binding, cpa, cpa_action_binding):
+def match_action_bindings(cpp, cpp_action_binding, cpa, cpa_action_binding, direction, other_parties):
+    acl_check(cpp_action_binding, cpp, other_parties)
+    match_payload_profiles(cpp, cpp_action_binding, cpa, cpa_action_binding)
     cpa_channel_id_element = cpa_action_binding.xpath('cppa:ChannelId', namespaces=_NSMAP)[0]
     for cpp_channel_id_element in cpp_action_binding.xpath('cppa:ChannelId', namespaces=_NSMAP):
         cpa_channel_id = cpa_channel_id_element.text
         cpp_channel_id = cpp_channel_id_element.text
         try:
-            match_channels(cpp, cpp_channel_id, cpa, cpa_channel_id)
+            match_channels(cpp, cpp_channel_id, cpa, cpa_channel_id, direction)
             logging.info('Matched CPA channel {} to CPP channel {}'.format(cpa_channel_id,
                                                                            cpp_channel_id))
-            return cpp_channel_id
-        except:
+        except MatchException:
             exception = str(traceback.format_exc())
             logging.info('Unable to match CPA channel {} to CPP channel {}: {}'.format(cpa_channel_id,
                                                                                        cpp_channel_id,
                                                                                        exception))
+        else:
+            return cpp_channel_id
+
     # if we're here, all matches raised exceptions
     exception = 'Unable to match CPA channel {} to any CPP channel'.format(cpa_channel_id)
     logging.info(exception)
     raise MatchException(exception)
 
-def match_channels(cpp, cpp_channel_id, cpa, cpa_channel_id):
+def match_payload_profiles(cpp, cpp_action_binding, cpa, cpa_action_binding):
+    cpp_profile_id = cpp_action_binding.get('payloadProfileId')
+    cpa_profile_id = cpa_action_binding.get('payloadProfileId')
+    logging.info('Matching payload profiles {} and {}'.format(cpp_profile_id, cpa_profile_id))
+    if cpp_profile_id == cpa_profile_id == None:
+        pass
+    else:
+        cpp_profile = cpp.xpath('cppa:PayloadProfile[@id="{}"]'.format(cpp_profile_id), namespaces=_NSMAP)[0]
+        cpa_profile = cpa.xpath('cppa:PayloadProfile[@id="{}"]'.format(cpa_profile_id), namespaces=_NSMAP)[0]
+        cpp_part_list = cpp_profile.xpath('child::cppa:PayloadPart', namespaces=_NSMAP)
+        cpa_part_list = cpa_profile.xpath('child::cppa:PayloadPart', namespaces=_NSMAP)
+        cpplen = len(cpp_part_list)
+        cpalen = len(cpa_part_list)
+        if cpplen != cpalen:
+            raise MatchException(
+                'Inconsistent number of payload parts {}: {}; {}: {}'.format(cpp_profile_id,
+                                                                                   cpplen,
+                                                                                   cpa_profile_id,
+                                                                                   cpalen)
+            )
+        else:
+            for c in range(0,cpplen):
+                cpp_part = cpp_part_list[c]
+                cpa_part = cpa_part_list[c]
+                match_payload_part(cpp, cpp_profile_id, cpa, cpa_profile_id, c, cpp_part, cpa_part)
+
+
+def match_payload_part(cpp, cpp_profile_id, cpa, cpa_profile_id, c, cpp_part, cpa_part):
+    logging.info('Matching payload part {} for {} and {}'.format(c, cpp_profile_id, cpa_profile_id))
+    for cpa_child_element in cpa_part:
+        if cpa_child_element.tag in [
+            cppa('PartName'), cppa('MIMEContentType'), cppa('Property')]:
+            match_simple_child_element(cpp_part, cpa_child_element)
+
+def match_channels(cpp, cpp_channel_id, cpa, cpa_channel_id, direction):
     logging.info('Checking CPA channel {} and CPP channel {}'.format(cpa_channel_id,
                                                                      cpp_channel_id))
     cpp_channel = cpp.xpath('*[@id="{}"]'.format(cpp_channel_id))[0]
     cpa_channel = cpa.xpath('*[@id="{}"]'.format(cpa_channel_id))[0]
-    assert_equal(cpp_channel.tag, cpa_channel.tag)
-    match_schema(cpp, cpp_channel, cpa, cpa_channel)
-    match_transports(cpp, cpp_channel, cpa, cpa_channel)
+    match_channel(cpp, cpp_channel, cpa, cpa_channel, direction)
 
-def match_transports(cpp, cpp_channel, cpa, cpa_channel):
+def match_channel(cpp, cpp_channel, cpa, cpa_channel, direction):
+    if cpa_channel.tag == cppa('DelegationChannel'):
+        if cpp_channel.tag == cppa('DelegationChannel'):
+            cpp_delegate_id = cpp_channel.xpath('cppa:PartyId/text()',
+                                            namespaces=_NSMAP)[0]
+            role_in_cpa, cpa_delegate_id_list = linked_delegate_party(cpa_channel, direction)
+            if len(cpa_delegate_id_list) == 0:
+                raise MatchException('No {}, {} in CPA'.format(role_in_cpa, cpp_delegate_id))
+            else:
+                cpa_delegate_id = cpa_delegate_id_list[0]
+                if cpp_delegate_id != cpa_delegate_id:
+                    raise MatchException('Party delegation mismatch, CPP {}, CPA {}'.format(cpa_delegate_id,
+                                                                                            cpp_delegate_id))
+        else:
+            role_in_cpa, cpa_delegate_id_list = linked_delegate_party(cpa_channel, direction)
+            if len(cpa_delegate_id_list) > 0:
+                raise MatchException('No DelegationChannel in CPP, but there is a {} in CPA'.format(
+                    cpa_delegate_id_list[0])
+                )
+
+    else:
+        assert_equal(cpp_channel.tag, cpa_channel.tag)
+        match_schema(cpp, cpp_channel, cpa, cpa_channel, direction)
+        match_transports(cpp, cpp_channel, cpa, cpa_channel, direction)
+
+def linked_delegate_party(channel, direction):
+    if direction == reverse:
+        return 'CounterPartyId', channel.xpath('cppa:CounterPartyId/text()', namespaces=_NSMAP)
+    else:
+        return 'PartyId', channel.xpath('cppa:PartyId/text()', namespaces=_NSMAP)
+
+def match_transports(cpp, cpp_channel, cpa, cpa_channel, direction):
     cpp_transport_id = cpp_channel.get('transport')
     cpa_transport_id = cpa_channel.get('transport')
+    logging.info('Checking transports {} and {}'.format(cpp_transport_id,
+                                                        cpa_transport_id))
     if cpp_transport_id != None and cpa_transport_id != None:
         cpp_transport = cpp.xpath('*[@id="{}"]'.format(cpp_transport_id))[0]
         cpa_transport = cpa.xpath('*[@id="{}"]'.format(cpa_transport_id))[0]
-        match_schema(cpp, cpp_transport, cpa, cpa_transport)
+        match_schema(cpp, cpp_transport, cpa, cpa_transport, direction)
 
 
-def match_schema(cpp, cpp_element, cpa, cpa_element):
+def match_schema(cpp, cpp_element, cpa, cpa_element, direction):
+    if cpp_element.tag != cpa_element.tag:
+        raise MatchException('Incompatible elements: {} {}'.format(cpp_element.tag,
+                                                                   cpa_element.tag))
     if cpa_element.tag in cppa3_content_model:
         # the element is a complex element
         if cpa_element.tag in _allowed_non_matching_complex_elements:
@@ -218,23 +315,84 @@ def match_schema(cpp, cpp_element, cpa, cpa_element):
                     else:
                         logging.info('Checking complex child element {}'.format(cpa_child_element.tag))
                         cpp_child_element = retrieve_child_element(cpa_child_element.tag, cpp_element)
-                        match_schema(cpp, cpp_child_element, cpa, cpa_child_element)
+                        match_schema(cpp, cpp_child_element, cpa, cpa_child_element, direction)
                         logging.info('Checked complex child element {}'.format(cpa_child_element.tag))
                 else:
                     if cpa_child_element.tag in _allowed_non_matching_simple_elements:
-                        logging.info('Skipping exceptional {}'.format(cpa_child_element.tag))
+                        logging.info('Skipping allowed non matching {}'.format(cpa_child_element.tag))
                     elif cpa_child_element.tag in _channel_referencing_elements:
                         logging.info('Processing channel reference {}'.format(cpa_child_element.tag))
-                        match_channel_reference_element(cpp, cpp_element, cpa, cpa_child_element)
+                        match_channel_reference_element(cpp, cpp_element, cpa, cpa_child_element, direction)
                     elif cpa_child_element.tag in _certificate_referencing_elements:
                         match_certificate_reference_element(cpp, cpp_element, cpa, cpa_child_element)
                     else:
-                        logging.error('Simple element:  {}'.format(cpa_child_element.tag))
                         match_simple_child_element(cpp_element, cpa_child_element)
+    else:
+        logging.debug('{} is not in CPPA3 content model'.format(cpa_element.tag))
+
+
+"""
+Authorization
+"""
+
+def cpa_other_party_ids(cpa, direction):
+    if direction == reverse:
+        xpath_expression = 'child::cppa:PartyInfo/cppa:PartyId'
+    else:
+        xpath_expression = 'child::cppa:CounterPartyInfo/cppa:PartyId'
+    parties = []
+    for party_id in cpa.xpath(xpath_expression,
+        namespaces=_NSMAP):
+        pid = party_id.text
+        pidtype = party_id.get('type')
+        if (pid, pidtype) not in parties:
+            parties.append((pid,pidtype))
+    logging.info('@@@ {} {}'.format(direction, str(parties)))
+    return parties
+
+def acl_check(cpp_element, cpp, other_parties):
+    cpp_allowed_party_list_id = cpp_element.get('allowed')
+    cpp_denied_party_list_id = cpp_element.get('denied')
+    if cpp_allowed_party_list_id == None:
+        pass
+    else:
+        cpp_allowed_parties = lookup_party_identifiers(cpp,
+                                                       cpp_allowed_party_list_id, [])
+        acl_allow_match(cpp_allowed_parties, other_parties)
+
+    if cpp_denied_party_list_id == None:
+        pass
+    else:
+        cpp_denied_parties = lookup_party_identifiers(cpp,
+                                                      cpp_denied_party_list_id, [])
+        acl_deny_match(cpp_denied_parties, other_parties)
+
+def acl_allow_match(allowed_party_ids, party_ids):
+    for (pid, ptype) in party_ids:
+        if (pid, ptype) not in allowed_party_ids:
+            raise_match_exception(
+                '{}, {} not in allowed list {}'.format(pid,
+                                                       ptype,
+                                                       allowed_party_ids))
+        else:
+            logging.debug('{}, {} found in allowed party list'.format(pid,
+                                                                      ptype))
+
+def acl_deny_match(denied_party_ids, party_ids):
+    for (pid, ptype) in party_ids:
+        if (pid, ptype) in denied_party_ids:
+            raise_match_exception(
+                '{}, {} in denied list {}'.format(pid,
+                                                  ptype,
+                                                  denied_party_ids))
+        else:
+            logging.debug('{}, {} not in denied party list'.format(pid,
+                                                                   ptype))
 
 
 def assert_equal(value_1, value_2):
-    if value_1 != value_2:
+    logging.debug('... {} {}'.format(value_1, value_2))
+    if str(value_1) != str(value_2):
         raise MatchException('{} != {}'.format(value_1, value_2))
 
 def retrieve_child_element(child_element_tag, parent_element):
@@ -270,7 +428,7 @@ def match_simple_child_element(parent_element, child_element):
     else:
         logging.info('Found one {}'.format(child_element.tag))
 
-def match_channel_reference_element(cpp, parent_element, cpa, child_element):
+def match_channel_reference_element(cpp, parent_element, cpa, child_element, direction):
     local_name = re.split('\}', child_element.tag)[1]
     candidate_matches = parent_element.xpath(
         'cppa:{}'.format(local_name),
@@ -286,7 +444,7 @@ def match_channel_reference_element(cpp, parent_element, cpa, child_element):
         logging.info('Matching referenced {} channels {} and {}'.format(child_element.tag,
                                                                         child_element.text,
                                                                         cpp_channel_id))
-        match_channels(cpp, cpp_channel_id, cpa, child_element.text)
+        match_channels(cpp, cpp_channel_id, cpa, child_element.text, direction)
 
 def match_certificate_reference_element(cpp, parent_element, cpa, child_element):
     local_name = re.split('\}', child_element.tag)[1]
@@ -345,6 +503,7 @@ def check_cpa_cert_with_cpp_trust_anchors(cpp, cpp_parent, cpa_certificate, loca
             trust_anchor_set = cpp.xpath('descendant::cppa:*[@id="{}"]'.format(trust_anchor_set_id),
                                          namespaces=_NSMAP)[0]
             rootfound = False
+            logging.info('Checking certificate against trust anchor {}'.format(trust_anchor_set_id))
             for anchor_certificate_ref in trust_anchor_set.xpath('cppa:AnchorCertificateRef',
                 namespaces=_NSMAP):
                 anchor_certificate_ref_id = anchor_certificate_ref.get('certId')
@@ -361,12 +520,13 @@ def check_cpa_cert_with_cpp_trust_anchors(cpp, cpp_parent, cpa_certificate, loca
                 if trusted_ca_certificate_x509 == cpa_x509_certificate_root:
                     rootfound = True
             if not rootfound:
+                logging.info('Certificate was not a referenced anchor certificate. Checking embedded certs')
                 for trusted_ca_certificate in trust_anchor_set.xpath(
                         'cppa:Certificate', namespaces=_NSMAP
                 ):
                     trusted_ca_certificate_x509 = remove_all_whitespace(
                         trusted_ca_certificate.xpath(
-                            'descendant::ds:X509Certificate', namespaces=_NSMAP
+                            'descendant::ds:X509Certificate/text()', namespaces=_NSMAP
                         )[-1]
                     )
                     if trusted_ca_certificate_x509 == cpa_x509_certificate_root:
@@ -434,3 +594,9 @@ def remove_all_whitespace(inputstring):
     pattern = re.compile(r'\s+')
     return re.sub(pattern, '', inputstring)
 
+def cppa(el):
+    return '{{{}}}{}'.format(_NSMAP['cppa'],el)
+
+def raise_match_exception(value):
+    logging.info('MatchException: {}'.format(value))
+    raise MatchException(value)
\ No newline at end of file
