@@ -3,7 +3,7 @@ __author__ = 'pvde'
 """
 Module to implement support for ChannelProfile in CPPA3
 
-A ChannelProfileConfig class
+A ChannelProfileHandler class
 
 """
 
