@@ -13,7 +13,7 @@ from copy import deepcopy
 
 NSMAP = {'cppa': 'http://docs.oasis-open.org/ebcore/ns/cppa/v3.0' }
 
-class CPPA3TestCase( unittest.TestCase ):
+class UnifyTestCase( unittest.TestCase ):
 
     def setUp(self):
         logging.getLogger('').handlers = []
